import java.util.*;
import javax.security.sasl.*;
import javax.security.auth.callback.*;

public class SASLCheck
{

	private static boolean mangle = false;
	private static int loopCount = 1;

	public static void main(String[] args) throws Exception
	{
		PrintSysInfo();
		//System.in.read();
		PrintFactories();
		for (int i = 0; i < args.length; i++)
		{
			try
			{
				for (int j = 0; j < loopCount; j++)
				{
					DoSaslTest(args[i]);
				}
			}
			catch (Exception ex)
			{
				System.out.println(ex.toString());
			}
		}
		//DoSaslTest("EXTERNAL");
		//DoSaslTest("PLAIN");
		//DoSaslTest("DIGEST-MD5");
		//DoSaslTest("CRAM-MD5");
		//DoSaslTest("WIERD");
		//DoSaslTest("NTLM");
		//DoSaslTest("GSSAPI");
		//System.out.println("Press enter to continue . . .");
		//System.in.read();
	}

	private static void PrintSysInfo()
	{
		System.out.println("Using:");
		System.out.println(System.getProperty("java.vendor"));
		System.out.println(System.getProperty("java.version"));
		System.out.println();
	}

	private static void PrintFactories() throws Exception
	{
		System.out.println("ClientFactories:");
		Enumeration<SaslClientFactory> clientList = Sasl.getSaslClientFactories();

		for (; clientList.hasMoreElements() ;)
		{

			SaslClientFactory sasl = (SaslClientFactory)clientList.nextElement();

			String[] names = sasl.getMechanismNames(null);

			System.out.println(" " + sasl.getClass() + " has " + names.length + " registered mechanism(s):");
			for (int i = 0; i < names.length; i++)
			{
				System.out.println("  " + (i+1) + ": " + names[i]);
			}

		}
		System.out.println();

		System.out.println("ServerFactories:");
		Enumeration<SaslServerFactory> serverList = Sasl.getSaslServerFactories();

		for (; serverList.hasMoreElements() ;)
		{

			SaslServerFactory sasl = (SaslServerFactory)serverList.nextElement();

			String[] names = sasl.getMechanismNames(null);

			System.out.println(" " + sasl.getClass() + " has " + names.length + " registered mechanism(s):");
			for (int i = 0; i < names.length; i++)
			{
				System.out.println("  " + (i+1) + ": " + names[i]);
			}

		}
		System.out.println();

	}

	private static void DoSaslTest(String mech) throws Exception
	{
		System.out.println("Testing: " + mech);
		mcount = 1;

		String authzid = "username";
		String protocol = "xmpp";
		String serverName = "theserver";
		Map<String,Object> props = new HashMap<String,Object>();
		props.put("com.sun.security.sasl.digest.realm", "realm");
		String myName = serverName;

		CallbackHandler callbackHandler = new ClientCallback();

		String[] mechanisms = new String[]{mech};
		SaslClient sc = Sasl.createSaslClient(mechanisms,
			authzid, protocol, serverName, props, callbackHandler);

		callbackHandler = new ServerCallback();

		SaslServer ss = Sasl.createSaslServer(mech,
			protocol, myName, props, callbackHandler);

		if (sc == null) System.out.println("Can't create client mech");
		if (ss == null) System.out.println("Can't create server mech");
		if ((sc == null) || (ss == null)) return;

		System.out.println("Client mechname: " + sc.getMechanismName());
		System.out.println("Server mechname: " + ss.getMechanismName());

		// Get optional initial response
		byte[] sasldata = new byte[] {};
		sasldata = (sc.hasInitialResponse() ? sc.evaluateChallenge(sasldata) : null);
		byte[] lastclient = sasldata;

		while (!ss.isComplete())
		{
			if (sasldata == null) sasldata = new byte[] {};
			if (mangle) sasldata = Mangle(sasldata);
			System.out.println("c2s: '" + new String(sasldata).replace('\007', ' ') + "'");

			sasldata = ss.evaluateResponse(sasldata);

			if (sasldata == null) sasldata = new byte[] {};
			if (mangle) sasldata = Mangle(sasldata);

			System.out.println("s2c: '" + new String(sasldata).replace('\007', ' ') + "'");

			if (!sc.isComplete())
			{
				sasldata = sc.evaluateChallenge(sasldata);

				if (lastclient == sasldata)
				{
					System.out.println("ABORT: client response was same twice in a row");
					return;
				}
				lastclient = sasldata;
			}

			if (ss.isComplete() && sc.isComplete())
			{
				System.out.println("SUCCESS: User is " + ss.getAuthorizationID());
				break;
			}
			//System.in.read();
		}

		sc = null;
		ss = null;
		System.gc();
		System.runFinalization();
	}

	static int mcount;
	private static byte[] Mangle(byte[] datain)
	{
		if ((mcount == 3) && (datain.length > 8))
		{
			datain[8] = 0;
		}
		mcount++;
		return datain;
	}
}

class ClientCallback implements CallbackHandler
{
	public void handle(javax.security.auth.callback.Callback[] callbacks) throws UnsupportedCallbackException
	{
		System.out.println(" Client handling " + callbacks.length + " callbacks");
		for (int i = 0; i < callbacks.length; i++)
		{
			System.out.println("  Callback: " + callbacks[i].getClass().getSimpleName());
			if (callbacks[i] instanceof RealmCallback)
			{
				((RealmCallback) callbacks[i]).setText("realm");
			}
			else if (callbacks[i] instanceof NameCallback)
			{
				((NameCallback) callbacks[i]).setName("username");
			}
			else if (callbacks[i] instanceof PasswordCallback)
			{
				((PasswordCallback) callbacks[i]).setPassword("passwrd1".toCharArray());
			}
			else
			{
				throw new UnsupportedCallbackException(callbacks[i], "Unrecognized Callback");
			}
		}
	}
}

class ServerCallback implements CallbackHandler
{
	public void handle(javax.security.auth.callback.Callback[] callbacks) throws UnsupportedCallbackException
	{
		System.out.println(" Server handling " + callbacks.length + " callbacks");
		String realm = null;
		String name = null;
		for (int i = 0; i < callbacks.length; i++)
		{
			System.out.println("  Callback: " + callbacks[i].getClass().getSimpleName());
			if (callbacks[i] instanceof RealmCallback)
			{
				realm = ((RealmCallback) callbacks[i]).getText();
				if (realm == null)
				{
					realm = ((RealmCallback) callbacks[i]).getDefaultText();
				}
				System.out.println("   RealmCallback: " + realm);
			}
			else if (callbacks[i] instanceof NameCallback)
			{
				name = ((NameCallback) callbacks[i]).getName();
				if (name == null)
				{
					name = ((NameCallback) callbacks[i]).getDefaultName();
				}
				System.out.println("   NameCallback: " + name);
			}
			else if (callbacks[i] instanceof PasswordCallback)
			{
				if (((PasswordCallback) callbacks[i]).getPassword() != null) {
					// The client has supplied the auth mech with a plaintext password,
					//  so we can use it in other authentication mechanisms.
					// Try and authenticate using the password supplied to us,
					//  if it's okay then we just succeed
					System.out.println("   PasswordCallback Client: " + new String(((PasswordCallback) callbacks[i]).getPassword()));
				} else {
                    // The client can't supply the auth mech with a plaintext password, so we have to.

				    // Get the password from the UserProvider. Some UserProviders may not support
				    // this operation
				    ((PasswordCallback) callbacks[i]).setPassword("passwrd1".toCharArray());

				    System.out.println("   PasswordCallback Server: " + new String(((PasswordCallback) callbacks[i]).getPassword()));
				}
			}
			else if (callbacks[i] instanceof AuthorizeCallback)
			{
				AuthorizeCallback authCallback = ((AuthorizeCallback) callbacks[i]);
				String authenId = authCallback.getAuthenticationID();
				String authorId = authCallback.getAuthorizationID();
				if ((authenId == null) || (authenId.equals(authorId) || (authenId.startsWith(authorId + "@"))))
				{
					authCallback.setAuthorized(true);
					authCallback.setAuthorizedID(authorId);
				}
				System.out.println("   AuthorizeCallback: authenId: " + authenId);
				System.out.println("   AuthorizeCallback: authorId: " + authorId);
			}
			else
			{
				throw new UnsupportedCallbackException(callbacks[i], "Unrecognized Callback");
			}
		}
	}
}
