package net.za.darkskies.security.sasl;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Map;
import javax.security.auth.callback.*;
import javax.security.sasl.*;

final class PlainServer
    implements SaslServer
{

	private CallbackHandler cbh;
	protected boolean completed;
	private String authorizationID;
	private static byte SEP = 0;

	PlainServer(CallbackHandler cbh)
        throws SaslException
    {
		this.cbh = cbh;
		completed = false;
	}

	public String getMechanismName()
	{
		return "PLAIN";
	}

	public boolean isComplete()
	{
		return completed;
	}

	public byte[] unwrap(byte incoming[], int offset, int len)
		throws SaslException
	{
		if (completed)
			throw new IllegalStateException("PLAIN supports neither integrity nor privacy");
		else
			throw new IllegalStateException("PLAIN authentication not completed");
	}

	public byte[] wrap(byte outgoing[], int offset, int len)
		throws SaslException
	{
		if (completed)
			throw new IllegalStateException("PLAIN supports neither integrity nor privacy");
		else
			throw new IllegalStateException("PLAIN authentication not completed");
	}

    public byte[] evaluateResponse(byte response[])
        throws SaslException
    {
        if(completed)
            throw new IllegalStateException("PLAIN authentication already completed");

		int s1 = 0;
		int s2 = 0;
		for (int i = 0; i < response.length; i++)
			if (response[i] == SEP)
			{
				s1 = i + 1;
				break;
			}
		for (int i = s1; i < response.length; i++)
			if (response[i] == SEP)
			{
				s2 = i + 1;
				break;
			}

		String authid;
		String name;
		String pswd;
		try
		{
			authid = new String(response, 0, s1 - 1, "UTF8");
			name = new String(response, s1, s2 - s1 - 1, "UTF8");
			pswd = new String(response, s2, response.length - s2, "UTF8");
		}
		catch (UnsupportedEncodingException ex)
		{
			throw new SaslException("UTF8 not available on platform", ex);
		}

		NameCallback ncb = new NameCallback("PLAIN authentication ID: ", name);
		PasswordCallback pcb = new PasswordCallback("PLAIN password:", false);
		pcb.setPassword(pswd.toCharArray());
		pswd = null;
		try 
		{
			cbh.handle(new Callback[] {ncb, pcb});
		}
		catch (UnsupportedCallbackException ex)
		{
			throw new SaslException("PLAIN authentication failed", ex);
		}
		catch (IOException ex)
		{
			throw new SaslException("PLAIN authentication failed", ex);
		}

		char[] password = pcb.getPassword();
		if (password == null || password.length == 0)
		{
			throw new SaslException((new StringBuilder()).append("PLAIN: username not found: ").append(name).toString());
		}

		if(password.length != response.length - s2)
		{
			throw new SaslException("Invalid response");
		}
		
		int j = 0;
		for (int i = s2; i < response.length; i++)
			if(password[j++] != response[i])
			{
				throw new SaslException("Invalid response");
			}

		AuthorizeCallback authorizecallback = new AuthorizeCallback(name, authid);
		try
		{
		cbh.handle(new Callback[] 
			{
				authorizecallback
			});
		}
		catch (UnsupportedCallbackException ex)
		{
			throw new SaslException("PLAIN authentication failed", ex);
		}
		catch (IOException ex)
		{
			throw new SaslException("PLAIN authentication failed", ex);
		}

		if(authorizecallback.isAuthorized()) 
		{
			completed = true;
			authorizationID = authorizecallback.getAuthorizedID();
		}
		else
			throw new SaslException((new StringBuilder()).append("PLAIN: ").append(name).append(" is not authorized to act as ").append(authid).toString());
		
		return null;
		
	}

	public Object getNegotiatedProperty(String s)
	{
		if (completed)
		{
			if(s.equals(Sasl.QOP))
				return "auth";
			else
				return null;
		} 
		else
		{
			throw new IllegalStateException("SSPI authentication not completed");
		}
	}

	public String getAuthorizationID()
    {
        if (completed)
            return authorizationID;
        else
            throw new IllegalStateException("PLAIN authentication not completed");
    }

	public void dispose()
		throws SaslException
	{
	}

}
